
"""
Provides :class:`EngineBase`.
"""

from __future__ import print_function

import os
import sys
import typing
from returnn.log import log
from returnn.pretrain import Pretrain
from returnn.util import basic as util


class EngineBase(object):
  """
  Base class for a backend engine, such as :class:`TFEngine.Engine`.
  """

  def __init__(self):
    self.epoch = 0
    self.pretrain = None  # type: typing.Optional[Pretrain]
    self.model_filename = None  # type: typing.Optional[str]

  @classmethod
  def config_get_final_epoch(cls, config):
    """
    :param returnn.config.Config config:
    :rtype: int
    """
    num_epochs = config.int('num_epochs', 5)
    if config.has("load_epoch"):
      num_epochs = max(num_epochs, config.int("load_epoch", 0))
    return num_epochs

  @classmethod
  def get_existing_models(cls, config):
    """
    :param returnn.config.Config config:
    :return: dict epoch -> model filename
    :rtype: dict[int,str]
    """
    model_filename = config.value('model', '')
    if not model_filename:
      return {}
    # Automatically search the filesystem for existing models.
    file_list = {}
    for epoch in range(1, cls.config_get_final_epoch(config) + 1):
      for is_pretrain in [False, True]:
        fn = cls.epoch_model_filename(model_filename, epoch, is_pretrain)
        if os.path.exists(fn):
          file_list[epoch] = fn
          break
        if util.BackendEngine.is_tensorflow_selected():
          if os.path.exists(fn + ".index"):
            file_list[epoch] = fn
            break
        elif util.BackendEngine.is_torch_selected():
          if os.path.exists(fn + ".pt"):
            file_list[epoch] = fn
            break
    return file_list

  @classmethod
  def get_epoch_model(cls, config):
    """
    :type config: returnn.config.Config
    :returns (epoch, modelFilename)
    :rtype: (int|None, str|None)
    """
    start_epoch_mode = config.value('start_epoch', 'auto')
    if start_epoch_mode == 'auto':
      start_epoch = None
    else:
      start_epoch = int(start_epoch_mode)
      assert start_epoch >= 1

    load_model_epoch_filename = util.get_checkpoint_filepattern(config.value('load', ''))
    if load_model_epoch_filename:
      assert os.path.exists(load_model_epoch_filename + util.get_model_filename_postfix())

    import_model_train_epoch1 = util.get_checkpoint_filepattern(config.value('import_model_train_epoch1', ''))
    if import_model_train_epoch1:
      assert os.path.exists(import_model_train_epoch1 + util.get_model_filename_postfix())

    existing_models = cls.get_existing_models(config)
    load_epoch = config.int("load_epoch", -1)
    if load_model_epoch_filename:
      if load_epoch <= 0:
        load_epoch = util.model_epoch_from_filename(load_model_epoch_filename)
    else:
      if load_epoch > 0:  # ignore if load_epoch == 0
        assert load_epoch in existing_models
        load_model_epoch_filename = existing_models[load_epoch]
        assert util.model_epoch_from_filename(load_model_epoch_filename) == load_epoch

    # Only use this when we don't train.
    # For training, we first consider existing models before we take the 'load' into account when in auto epoch mode.
    # In all other cases, we use the model specified by 'load'.
    if load_model_epoch_filename and (config.value('task', 'train') != 'train' or start_epoch is not None):
      if config.value('task', 'train') == 'train' and start_epoch is not None:
        # Ignore the epoch. To keep it consistent with the case below.
        epoch = None
      else:
        epoch = load_epoch
      epoch_model = (epoch, load_model_epoch_filename)

    # In case of training, always first consider existing models.
    # This is because we reran RETURNN training, we usually don't want to train from scratch
    # but resume where we stopped last time.
    elif existing_models:
      epoch_model = sorted(existing_models.items())[-1]
      if load_model_epoch_filename:
        print("note: there is a 'load' which we ignore because of existing model", file=log.v4)

    elif config.value('task', 'train') == 'train' and import_model_train_epoch1 and start_epoch in [None, 1]:
      epoch_model = (0, import_model_train_epoch1)

    # Now, consider this also in the case when we train, as an initial model import.
    elif load_model_epoch_filename:
      # Don't use the model epoch as the start epoch in training.
      # We use this as an import for training.
      epoch_model = (load_epoch, load_model_epoch_filename)

    else:
      epoch_model = (None, None)

    if start_epoch == 1:
      if epoch_model[0]:  # existing model
        print("warning: there is an existing model: %s" % (epoch_model,), file=log.v4)
        epoch_model = (None, None)
    elif (start_epoch or 0) > 1:
      if epoch_model[0]:
        if epoch_model[0] != start_epoch - 1:
          print("warning: start_epoch %i but there is %s" % (start_epoch, epoch_model), file=log.v4)
        epoch_model = start_epoch - 1, existing_models[start_epoch - 1]

    return epoch_model

  @classmethod
  def get_train_start_epoch_batch(cls, config):
    """
    We will always automatically determine the best start (epoch,batch) tuple
    based on existing model files.
    This ensures that the files are present and enforces that there are
    no old outdated files which should be ignored.
    Note that epochs start at idx 1 and batches at idx 0.
    :type config: returnn.config.Config
    :returns (epoch,batch)
    :rtype (int,int)
    """
    start_batch_mode = config.value('start_batch', 'auto')
    if start_batch_mode == 'auto':
      start_batch_config = None
    else:
      start_batch_config = int(start_batch_mode)
    last_epoch, _ = cls.get_epoch_model(config)
    if last_epoch is None:
      start_epoch = 1
      start_batch = start_batch_config or 0
    elif start_batch_config is not None:
      # We specified a start batch. Stay in the same epoch, use that start batch.
      start_epoch = last_epoch
      start_batch = start_batch_config
    else:
      # Start with next epoch.
      start_epoch = last_epoch + 1
      start_batch = 0
    return start_epoch, start_batch

  @classmethod
  def epoch_model_filename(cls, model_filename, epoch, is_pretrain):
    """
    :type model_filename: str
    :type epoch: int
    :type is_pretrain: bool
    :rtype: str
    """
    if sys.platform == "win32" and model_filename.startswith("/tmp/"):
      import tempfile
      model_filename = tempfile.gettempdir() + model_filename[len("/tmp"):]
    return model_filename + (".pretrain" if is_pretrain else "") + ".%03d" % epoch

  def get_epoch_model_filename(self, epoch=None):
    """
    :param int|None epoch:
    :return: filename, excluding TF specific postfix
    :rtype: str
    """
    if not epoch:
      epoch = self.epoch
    return self.epoch_model_filename(self.model_filename, epoch, self.is_pretrain_epoch(epoch=epoch))

  def get_epoch_str(self):
    """
    :return: e.g. "epoch 3", or "pretrain epoch 5"
    :rtype: str
    """
    return ("pretrain " if self.is_pretrain_epoch() else "") + "epoch %s" % self.epoch

  def is_pretrain_epoch(self, epoch=None):
    """
    :param int|None epoch:
    :return: whether this epoch is covered by the pretrain logic
    :rtype: bool
    """
    if not epoch:
      epoch = self.epoch
    return self.pretrain and epoch <= self.pretrain.get_train_num_epochs()

  def is_first_epoch_after_pretrain(self):
    """
    :return: whether the current epoch is the first epoch right after pretraining
    :rtype: bool
    """
    return self.pretrain and self.epoch == self.pretrain.get_train_num_epochs() + 1




"""
This module contains pretrain related code.
This is independent from the backend (TF or Theano, etc).
"""

from __future__ import print_function
from __future__ import absolute_import

from returnn.util.basic import unicode, long


class WrapEpochValue:
  """
  Use this wrapper if you want to define some value in your network
  which depends on the pretrain epoch.
  This is going to be part in your network description dict.
  """
  def __init__(self, func):
    """
    :param ((epoch: int) -> object) func: function which should accept one kwd-arg 'epoch'
    """
    self.func = func

  def get_value(self, epoch):
    """
    :param int epoch:
    :return: anything, whatever self.func returns
    :rtype: object
    """
    return self.func(epoch=epoch)


def find_pretrain_wrap_values(net_json):
  """
  See also :func:`Pretrain._resolve_wrapped_values`.
  Recursively goes through dicts, tuples and lists.
  This is a simple check to see if this is needed,
  i.e. if there are any :class:`WrapEpochValue` used.

  :param dict[str] net_json: network dict
  :return: whether there is some :class:`WrapEpochValue` in it
  :rtype: bool
  """
  assert isinstance(net_json, dict)

  def _check(d):
    if isinstance(d, WrapEpochValue):
      return True
    if isinstance(d, dict):
      for k, v in sorted(d.items()):
        if _check(v):
          return True
    if isinstance(d, (tuple, list)):
      for v in d:
        if _check(v):
          return True
    return False

  return _check(net_json)


class Pretrain:
  """
  Start with 1 hidden layers up to N hidden layers -> N pretrain steps -> N epochs (with repetitions == 1).
  The first hidden layer is the input layer.
  This works for generic network constructions. See _construct_epoch().
  """
  # Note: If we want to add other pretraining schemes, make this a base class.

  def __init__(self, original_network_json, network_init_args=None,
               copy_param_mode=None, copy_output_layer=None, greedy=None,
               repetitions=None,
               construction_algo="from_output", output_layers=("output",), input_layers=("data",)):
    """
    :type original_network_json: dict[str]
    :param dict[str]|None network_init_args: additional args we use for LayerNetwork.from_json().
      must have n_in, n_out. (for Theano only, thus optional now)
    :param str copy_param_mode:
    :param bool|str copy_output_layer: whether to copy the output layer params from last epoch or reinit
    :param bool greedy: if True, only train output+last layer, otherwise train all
    :param None | int | list[int] | dict repetitions: how often to repeat certain pretrain steps. default is one epoch.
      It can also be a dict, with keys like 'default' and 'final'. See code below.
    :param str|callable construction_algo: e.g. "from_output"
    :param list[str]|tuple[str] output_layers: used for construction
    :param list[str]|tuple[str] input_layers: used for construction
    """
    assert copy_param_mode in [None, "ifpossible", "subset", "reset"]
    if copy_output_layer is None:
      copy_output_layer = copy_param_mode
    if copy_output_layer is None:
      copy_output_layer = "ifpossible"
    if copy_output_layer == "reset":
      copy_output_layer = False
    if copy_output_layer:
      assert copy_output_layer is True or copy_output_layer in ["ifpossible", "subset"]
    self.copy_param_mode = copy_param_mode
    self.copy_output_layer = copy_output_layer
    if greedy is None:
      greedy = False
    self.greedy = greedy
    self.network_init_args = network_init_args
    self._original_network_json = original_network_json
    self._construction_algo = construction_algo
    self._input_layers = input_layers
    self._output_layers = output_layers
    if construction_algo == "from_input":
      self._construct_epochs_from_input()
    elif construction_algo == "from_output":
      self._construct_epochs_from_output()
    elif callable(construction_algo):
      self._construct_epochs_custom(construction_algo)
    elif construction_algo == "no_network_modifications":
      self._construct_epochs_no_network_modifications()
    else:
      raise Exception("invalid construction_algo %r" % construction_algo)
    if not callable(construction_algo):  # if callable, trust the user
      self._remove_non_trainable_added_only()
    if not repetitions:
      repetitions = 1
    if isinstance(repetitions, dict):
      rep_dict = repetitions.copy()
      default_rep = rep_dict.pop('default', 1)
      repetitions = [default_rep] * len(self._step_net_jsons)
      for k, v in sorted(rep_dict.items()):
        if k == "final":
          k = len(self._step_net_jsons) - 1
        repetitions[k] = v
    else:
      if not isinstance(repetitions, list):
        assert isinstance(repetitions, (int, long))
        repetitions = [repetitions]
      assert isinstance(repetitions, list)
      assert 0 < len(repetitions) <= len(self._step_net_jsons)
      if len(repetitions) < len(self._step_net_jsons):
        repetitions = repetitions + [repetitions[-1]] * (len(self._step_net_jsons) - len(repetitions))
    assert len(repetitions) == len(self._step_net_jsons)
    for i, net_dict in enumerate(self._step_net_jsons):
      if "#repetition" in net_dict:
        repetitions[i] = net_dict.pop("#repetition")
    self.repetitions = repetitions
    self._make_repetitions()
    self._resolve_wrapped_values()

  def _remove_non_trainable_added_only(self):
    """
    If from one epoch to the next, only non-trainable layers were added, remove this pretrain epoch.
    Output layers are ignored.
    Also handles first epoch.
    """
    assert self._step_net_jsons
    old_net_jsons = self._step_net_jsons
    self._step_net_jsons = []
    # -1 will be the empty net. Until one before final, which we will always add.
    for i in range(-1, len(old_net_jsons) - 2):
      if i == -1:
        net1, net2 = {}, old_net_jsons[0]
      else:
        net1, net2 = old_net_jsons[i:i+2]
      assert isinstance(net1, dict)
      assert isinstance(net2, dict)
      for layer_name in sorted(net1.keys()):
        assert layer_name in net2
      have_new_trainable = False
      for layer_name in sorted(net2.keys()):
        if self._is_layer_output(net2, layer_name):
          continue  # ignore output layers
        if layer_name in net1:
          continue  # already had before
        if net2[layer_name].get("trainable", True):
          have_new_trainable = True
          break
      if have_new_trainable:
        self._step_net_jsons.append(net2)
    # Always add final net.
    self._step_net_jsons.append(old_net_jsons[-1])

  def _make_repetitions(self):
    assert len(self.repetitions) == len(self._step_net_jsons)
    from copy import deepcopy
    old_net_jsons = self._step_net_jsons
    self._step_net_jsons = []
    for n_rep, net_json in zip(self.repetitions, old_net_jsons):
      for _ in range(n_rep):
        self._step_net_jsons.append(deepcopy(net_json))

  def _resolve_wrapped_values(self):
    """
    Resolves any :class:`WrapEpochValue` in the net dicts.
    Recursively goes through dicts, tuples and lists.
    See also :func:`find_pretrain_wrap_values`.
    """

    # noinspection PyShadowingNames
    def _check_dict(d, epoch, depth=0):
      for k, v in sorted(d.items()):
        if depth <= 1:  # 0 - layers, 1 - layer opts
          assert isinstance(k, (str, unicode))
        d[k] = _check(v, epoch=epoch, depth=depth + 1)

    # noinspection PyShadowingNames
    def _check(v, epoch, depth):
      """
      :param WrapEpochValue|tuple|list|dict|T v:
      :param int epoch:
      :param int depth:
      :rtype: T
      """
      if isinstance(v, WrapEpochValue):
        return v.get_value(epoch=epoch)
      if isinstance(v, (tuple, list)):
        if not any([isinstance(x, WrapEpochValue) for x in v]):
          return v
        return type(v)([_check(x, epoch=epoch, depth=depth + 1) for x in v])
      if isinstance(v, dict):
        _check_dict(v, epoch=epoch, depth=depth)
        return v
      return v

    for i, net_json in enumerate(self._step_net_jsons):
      epoch = i + 1
      _check_dict(net_json, epoch=epoch)

  # noinspection PyMethodMayBeStatic
  def _find_layer_descendants(self, json, sources):
    ls = []
    for other_layer_name, other_layer in sorted(json.items()):
      if other_layer_name in ls:
        continue
      other_sources = other_layer.get("from", ["data"])
      for src in sources:
        if src in other_sources:
          ls.append(other_layer_name)
          break
    return ls

  def _is_layer_output(self, json, layer_name):
    if layer_name in self._output_layers:
      return True
    if json[layer_name]["class"] == "softmax":
      return True
    if "target" in json[layer_name]:
      return True
    return False

  def _find_layer_outputs(self, json, sources):
    outs = []
    visited = set()
    while sources:
      visited.update(sources)
      for src in sources:
        if src in outs:
          continue
        if self._is_layer_output(json, src):
          outs.append(src)
      sources = self._find_layer_descendants(self._original_network_json, sources)
      for v in visited:
        if v in sources:
          sources.remove(v)
    return outs

  def _find_existing_inputs(self, json, layer_name, _collected=None, _visited=None):
    if _collected is None:
      _collected = []
    if _visited is None:
      _visited = {layer_name: None}
    sources = self._original_network_json[layer_name].get("from", ["data"])
    for src in sources:
      if src in json or src == "data":
        if src not in _collected:
          _collected.append(src)
      else:
        if src not in _visited:
          _visited[src] = layer_name
          self._find_existing_inputs(json=json, layer_name=src, _collected=_collected, _visited=_visited)
    return _collected

  def _construct_next_epoch_from_input(self, num_steps):
    """
    First find all layers which have data as input.
    Then expand from those layers.
    """
    from copy import deepcopy
    new_net = {}
    sources = ["data"]
    # Keep track of other layers which need to be added to make it complete.
    needed = set()

    # noinspection PyShadowingNames
    def update_needed(layer_name):
      """
      :param str layer_name:
      """
      needed.update(set(new_net[layer_name].get("from", ["data"])).difference(list(new_net.keys()) + ["data"]))
    # First search for non-trainable layers (like input windows or so).
    # You must specify "trainable": False in the layer at the moment.
    while True:
      descendants = self._find_layer_descendants(self._original_network_json, sources)
      added_something = False
      for layer_name in descendants:
        if layer_name in new_net:
          continue
        if self._original_network_json[layer_name].get("trainable", True):
          continue
        if layer_name in needed:
          needed.remove(layer_name)
        added_something = True
        sources.append(layer_name)
        new_net[layer_name] = deepcopy(self._original_network_json[layer_name])
        update_needed(layer_name)
      if not added_something:
        break
    # First do a search of depth `num_steps` through the net.
    for _ in range(num_steps):
      descendants = self._find_layer_descendants(self._original_network_json, sources)
      sources = []
      for layer_name in descendants:
        if layer_name in new_net:
          continue
        if layer_name in needed:
          needed.remove(layer_name)
        sources.append(layer_name)
        new_net[layer_name] = deepcopy(self._original_network_json[layer_name])
        update_needed(layer_name)
      if not sources:  # This means we reached the end.
        return False
    # Add all output layers.
    for layer_name in sorted(self._original_network_json.keys()):
      if layer_name in new_net:
        continue
      if not self._is_layer_output(self._original_network_json, layer_name):
        continue
      if layer_name in needed:
        needed.remove(layer_name)
      new_net[layer_name] = deepcopy(self._original_network_json[layer_name])
      update_needed(layer_name)
    if not needed:  # Nothing needed anymore, i.e. no missing layers, i.e. we arrived at the final network topology.
      return False
    # Now fill in all missing ones.
    for layer_name in sorted(new_net.keys()):
      sources = new_net[layer_name].get("from", ["data"])
      sources2 = self._find_existing_inputs(new_net, layer_name)
      if sources != sources2:
        if "data" in sources2:
          sources2.remove("data")
        new_net[layer_name]["from"] = sources2
    self._step_net_jsons.append(new_net)
    return True

  def _construct_epochs_from_input(self):
    self._step_net_jsons = []
    num_steps = 1
    while self._construct_next_epoch_from_input(num_steps):
      num_steps += 1
    # Just add the original net at the end.
    self._step_net_jsons.append(self._original_network_json)

  def _construct_new_epoch_from_output(self):
    """
    We start from the most simple network which we have constructed so far,
    and try to construct an even simpler network.
    """
    from copy import deepcopy
    new_json = deepcopy(self._step_net_jsons[0])
    while True:
      for out_layer_name in self._output_layers:
        assert out_layer_name in new_json
      # From the sources of the output layer, collect all their sources.
      # Then remove the direct output sources and replace them with the indirect sources.
      new_sources = set()
      deleted_sources = set()
      for out_layer_name in self._output_layers:
        for source in new_json[out_layer_name]["from"]:
          # Except for data sources. Just keep them.
          if source in self._input_layers:
            new_sources.add(source)
          else:
            assert source in new_json, "error %r, n: %i, last: %s" % (
              source, len(self._step_net_jsons), self._step_net_jsons[0])
            new_sources.update(new_json[source].get("from", ["data"]))
            del new_json[source]
            deleted_sources.add(source)
      # Check if anything changed.
      # This is e.g. not the case if the only source was data.
      if list(sorted(new_sources)) == list(sorted(set(sum(
            [new_json[name]["from"] for name in self._output_layers], [])))):
        return False
      for out_layer_name in self._output_layers:
        new_json[out_layer_name]["from"] = list(sorted(new_sources))
      # If we have data input, it likely means that the input dimension
      # for the output layer would change. Just avoid that for now.
      if new_sources.intersection(set(self._input_layers)):
        # Try again.
        continue
      # If all deleted sources were non-trainable, skip this.
      if all(not self._original_network_json[del_source].get("trainable", True) for del_source in deleted_sources):
        # Try again.
        continue
      self._step_net_jsons = [new_json] + self._step_net_jsons
      return True

  def _construct_epochs_from_output(self):
    self._step_net_jsons = [self._original_network_json]
    while self._construct_new_epoch_from_output():
      pass

  def _construct_epochs_custom(self, func):
    """
    :param ((idx: int, net_dict: dict[str,dict[str]]) -> dict[str,dict[str]]|None) func:
      ``func`` can work inplace on net_dict and should then return it.
      If ``None`` is returned, it will stop with the construction.
      The original network will always be added at the end.
    """
    from copy import deepcopy
    self._step_net_jsons = []
    idx = 0
    while True:
      d = func(idx=idx, net_dict=deepcopy(self._original_network_json))
      if not d:
        break
      self._step_net_jsons.append(d)
      idx += 1
    self._step_net_jsons.append(self._original_network_json)

  def _construct_epochs_no_network_modifications(self):
    self._step_net_jsons = [self._original_network_json]

  # -------------- Public interface

  def __str__(self):
    parts = [
      "Pretrain construction algo %r," % self._construction_algo,
      "number of pretrain epochs: %i" % self.get_train_num_epochs()]
    rep_set = set(self.repetitions)
    if rep_set != {1}:
      if len(rep_set) == 1:
        parts.append("(repetitions: %i)" % self.repetitions[0])
      else:
        parts.append("(repetitions: %r)" % self.repetitions)
    return " ".join(parts)

  def get_train_num_epochs(self):
    """
    :rtype: int
    """
    return len(self._step_net_jsons)

  def get_final_network_json(self):
    """
    :rtype: dict[str,dict[str]]
    """
    return self._step_net_jsons[-1]

  def get_network_json_for_epoch(self, epoch):
    """
    :param int epoch: starting at 1
    :rtype: dict[str]
    """
    assert epoch >= 1
    if epoch > len(self._step_net_jsons):
      epoch = len(self._step_net_jsons)  # take the last, which is the original
    return self._step_net_jsons[epoch - 1]

  def get_train_param_args_for_epoch(self, epoch):
    """
    :type epoch: int
    :returns the kwargs for LayerNetwork.set_train_params, i.e. which params to train.
    :rtype: dict[str]
    """
    if not self.greedy:
      return {}  # This implies all available args.
    if epoch == 1:
      return {}  # This implies all available args.
    raise NotImplementedError("This feature was removed with dropped Theano support")


def pretrain_from_config(config):
  """
  :type config: returnn.config.Config
  :rtype: Pretrain | None
  """
  from returnn.config import network_json_from_config
  pretrain_type = config.bool_or_other("pretrain", None)
  if pretrain_type == "default" or (isinstance(pretrain_type, dict) and pretrain_type) or pretrain_type is True:
    network_init_args = None
    original_network_json = network_json_from_config(config)
    opts = config.get_of_type("pretrain", dict, {})
    if config.has("pretrain_copy_output_layer"):
      opts.setdefault("copy_output_layer", config.bool_or_other("pretrain_copy_output_layer", "ifpossible"))
    if config.has("pretrain_greedy"):
      opts.setdefault("greedy", config.bool("pretrain_greedy", None))
    if config.has("pretrain_repetitions"):
      if config.is_typed("pretrain_repetitions"):
        opts.setdefault("repetitions", config.typed_value("pretrain_repetitions"))
      else:
        opts.setdefault("repetitions", config.int_list("pretrain_repetitions", None))
    if config.has("pretrain_construction_algo"):
      opts.setdefault("construction_algo", config.value("pretrain_construction_algo", None))
    return Pretrain(original_network_json=original_network_json, network_init_args=network_init_args, **opts)
  elif not pretrain_type:
    return None
  else:
    raise Exception("unknown pretrain type: %s" % pretrain_type)


def demo():
  """
  Will print out the different network topologies of the specified pretraining scheme.
  """
  import returnn.util.better_exchook
  returnn.util.better_exchook.install()
  import returnn.__main__ as rnn
  import argparse
  from returnn.util.basic import obj_diff_str
  arg_parser = argparse.ArgumentParser()
  arg_parser.add_argument("config")
  arg_parser.add_argument("--diff", action="store_true", help="show diff only")
  arg_parser.add_argument('other_returnn_args', nargs=argparse.REMAINDER, help="config updates or so")
  args = arg_parser.parse_args()
  rnn.init_config(
    config_filename=args.config,
    command_line_options=args.other_returnn_args,
    extra_updates={"log": []})
  # noinspection PyProtectedMember
  rnn.config._hack_value_reading_debug()
  rnn.init_log()
  if not rnn.config.value("pretrain", ""):
    print("config option 'pretrain' not set, will set it for this demo to 'default'")
    rnn.config.set("pretrain", "default")
  pretrain = pretrain_from_config(rnn.config)
  print("pretrain: %s" % pretrain)
  num_pretrain_epochs = pretrain.get_train_num_epochs()
  last_net_json = None
  from pprint import pprint
  for epoch in range(1, 1 + num_pretrain_epochs):
    print("epoch %i (of %i) network json:" % (epoch, num_pretrain_epochs))
    net_json = pretrain.get_network_json_for_epoch(epoch)
    if args.diff:
      if last_net_json is not None:
        print(obj_diff_str(last_net_json, net_json))
      else:
        print("(initial)")
    else:
      pprint(net_json)
    last_net_json = net_json
  print("done.")


if __name__ == "__main__":
  import sys
  import os
  __package__ = "returnn"  # pylint: disable=redefined-builtin
  my_dir = os.path.dirname(os.path.abspath(__file__))
  sys.path.insert(0, os.path.dirname(my_dir))
  if "" in sys.path:
    sys.path.remove("")
  if my_dir in sys.path:
    sys.path.remove(my_dir)
  import returnn
  sys.modules["returnn.pretrain"] = sys.modules["__main__"]
  try:
    demo()
  except BrokenPipeError:
    print("BrokenPipeError")
    sys.exit(1)




from __future__ import print_function
import _setup_test_env  # noqa
from nose.tools import assert_equal, assert_in, assert_not_in
from returnn.pretrain import pretrain_from_config
from returnn.config import Config


config_dict = {
  "pretrain": "default",
  "num_inputs": 40,
  "num_outputs": 4498,
}


net_dict = {
  "hidden_0": {"class": "linear", "n_out": 7, "dropout": 0.1, "activation": "relu"},
  "hidden_1": {"class": "linear", "n_out": 8, "dropout": 0.1, "activation": "relu",
               "from": ["hidden_0"]},
  "output":   {"class": "softmax", "loss": "ce", "from": ["hidden_1"]}
}


net_dict2 = {
  "lstm0_fw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": False},
  "lstm0_bw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": True},
  "lstm1_fw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": False,
               "from": ["lstm0_fw", "lstm0_bw"]},
  "lstm1_bw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": True,
               "from": ["lstm0_fw", "lstm0_bw"]},
  "lstm2_fw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": False,
               "from": ["lstm1_fw", "lstm1_bw"]},
  "lstm2_bw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": True,
               "from": ["lstm1_fw", "lstm1_bw"]},
  "output":   {"class": "softmax", "loss": "ce", "from": ["lstm2_fw", "lstm2_bw"]}
}


def test_config_net_dict1():
  config = Config()
  config.update(config_dict)
  config.typed_dict["network"] = net_dict
  pretrain = pretrain_from_config(config)
  assert_equal(pretrain.get_train_num_epochs(), 2)
  net1_json = pretrain.get_network_json_for_epoch(1)
  net2_json = pretrain.get_network_json_for_epoch(2)
  net3_json = pretrain.get_network_json_for_epoch(3)
  assert_in("hidden_0", net1_json)
  assert_not_in("hidden_1", net1_json)
  assert_in("hidden_0", net2_json)
  assert_in("hidden_1", net2_json)
  assert_equal(net2_json, net3_json)


def test_config_net_dict2():
  config = Config()
  config.update(config_dict)
  config.typed_dict["network"] = net_dict2
  pretrain = pretrain_from_config(config)
  assert_equal(pretrain.get_train_num_epochs(), 3)




# -*- coding: utf8 -*-

import sys
import os

import _setup_test_env  # noqa
sys.path += ["tools"]

from hdf_dump import *
import os
from returnn.log import log
import tempfile
from returnn.datasets.generating import DummyDataset
from returnn.datasets.hdf import HDFDataset
from returnn.util.basic import DictAsObj
import unittest
from returnn.util import better_exchook


options = {
  "epoch": 1,
  "start_seq": 0,
  "end_seq": float("inf")
}


def test_hdf_dataset_init():
  hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-init")
  hdf_dataset_init(hdf_filename)
  assert os.path.exists(hdf_filename)
  os.remove(hdf_filename)


def test_hdf_create():
  hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-create")
  hdf_dataset = hdf_dataset_init(hdf_filename)
  assert os.path.exists(hdf_filename)

  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
  dataset.init_seq_order(epoch=1)

  hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
  hdf_close(hdf_dataset)

  os.remove(hdf_filename)


def test_hdf_create_and_load():
  hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-load")
  hdf_dataset = hdf_dataset_init(hdf_filename)
  assert os.path.exists(hdf_filename)

  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
  dataset.init_seq_order(epoch=1)

  hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
  hdf_close(hdf_dataset)

  loaded_dataset = HDFDataset()
  loaded_dataset.add_file(hdf_filename)

  os.remove(hdf_filename)


def test_hdf_create_unicode_labels():
  hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-create")
  hdf_dataset = hdf_dataset_init(hdf_filename)
  assert os.path.exists(hdf_filename)

  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
  assert "classes" in dataset.get_target_list()
  dataset.labels["classes"] = ['’', 'ä', 'x']  # have some Unicode chars here
  dataset.init_seq_order(epoch=1)

  hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
  hdf_close(hdf_dataset)

  os.remove(hdf_filename)


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute



# coding: utf8
from __future__ import unicode_literals

import sys
import os
import unittest
import tempfile
import shutil
import gzip
import pickle

from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import raises

import _setup_test_env  # noqa
from returnn.util import better_exchook
from returnn.datasets.lm import TranslationDataset, TranslationFactorsDataset
from returnn.util.basic import init_thread_join_hack

better_exchook.install()
better_exchook.replace_traceback_format_tb()
init_thread_join_hack()


dummy_source_text = ("This is some example text.\n"
                     "It is used to test the translation dataset.\n"
                     "We will write it into a temporary file.\n")

dummy_target_text = ("Das ist ein Beispieltext.\n"
                     "Er wird zum Testen des TranslationDatasets benutzt.\n"
                     "Wir werden ihn in eine temporäre Datei schreiben.\n")


def create_vocabulary(text):
  """
  :param str text: any natural text
  :return: mapping of words in the text to ids, as well as the inverse mapping
  :rtype: (dict[str, int], dict[int, str])
  """
  vocabulary = {word: index for index, word in enumerate(set(text.strip().split()))}
  inverse_vocabulary = {index: word for word, index in vocabulary.items()}

  return vocabulary, inverse_vocabulary


def word_ids_to_sentence(word_ids, vocabulary):
  """
  :param list[int] word_ids:
  :param dict[int, str] vocabulary: mapping from word ids to words
  :return: concatenation of all words
  :rtype: str
  """
  words = [vocabulary[word_id] for word_id in word_ids]
  return ' '.join(words)


def test_translation_dataset():
  """
  Checks whether a dummy translation dataset can be read and whether the returned word indices are correct.
  We create the necessary corpus and vocabulary files on the fly.
  """

  dummy_dataset = tempfile.mkdtemp()
  source_file_name = os.path.join(dummy_dataset, "source.test.gz")  # testing both zipped and unzipped
  target_file_name = os.path.join(dummy_dataset, "target.test")

  with gzip.open(source_file_name, "wb") as source_file:  # writing in binary format works both in Python 2 and 3
    source_file.write(dummy_source_text.encode("utf-8"))

  with open(target_file_name, "wb") as target_file:
    target_file.write(dummy_target_text.encode("utf-8"))

  for postfix in ["", " </S>"]:  # test with and without postfix

    # Replace one word by <UNK>. This way it will not appear in the vocabulary (and <UNK> is added to the vocabulary).
    # We will test below whether this word is assigned the unknown id by checking whether the reconstruction also
    # contains <UNK>. Note, that the input file is already written and contains the original word.
    dummy_target_text_with_unk = dummy_target_text.replace("TranslationDatasets", "<UNK>")

    # Append postfix just to have it in the vocabulary
    source_vocabulary, inverse_source_vocabulary = create_vocabulary(dummy_source_text + postfix)
    target_vocabulary, inverse_target_vocabulary = create_vocabulary(dummy_target_text_with_unk + postfix)

    source_vocabulary_file_name = os.path.join(dummy_dataset, "source.vocab.pkl")
    target_vocabulary_file_name = os.path.join(dummy_dataset, "target.vocab.pkl")

    with open(source_vocabulary_file_name, "wb") as source_vocabulary_file:
      pickle.dump(source_vocabulary, source_vocabulary_file)

    with open(target_vocabulary_file_name, "wb") as target_vocabulary_file:
      pickle.dump(target_vocabulary, target_vocabulary_file)

    translation_dataset = TranslationDataset(
      path=dummy_dataset, file_postfix="test", source_postfix=postfix, target_postfix=postfix,
      unknown_label={"classes": "<UNK>"})
    translation_dataset.init_seq_order()
    translation_dataset.load_seqs(0, 10)

    num_seqs = len(dummy_source_text.splitlines())
    assert_equal(translation_dataset.num_seqs, num_seqs)

    # Reconstruct the sentences from the word ids and compare with input.
    for sequence_index in range(num_seqs):
      source_word_ids = translation_dataset.get_data(sequence_index, "data")
      source_sentence = word_ids_to_sentence(source_word_ids, inverse_source_vocabulary)
      assert_equal(source_sentence, dummy_source_text.splitlines()[sequence_index] + postfix)

      target_word_ids = translation_dataset.get_data(sequence_index, "classes")
      target_sentence = word_ids_to_sentence(target_word_ids, inverse_target_vocabulary)
      assert_equal(target_sentence, dummy_target_text_with_unk.splitlines()[sequence_index] + postfix)

  shutil.rmtree(dummy_dataset)


num_source_factors = 2
dummy_source_text_factor_0 = ("This is some example text.\n"
                              "The factors here have no meaning\n")
dummy_source_text_factor_1 = ("a b c d e\n"
                              "a b c d e f\n")
dummy_source_text_factored_format = ("This|a is|b some|c example|d text.|e\n"
                                     "The|a factors|b here|c have|d no|e meaning|f\n")

num_target_factors = 3
dummy_target_text_factor_0 = ("Das ist ein Beispieltext.\n"
                              "Die Factors hier haben keinen Sinn.\n")
dummy_target_text_factor_1 = ("a b c d\n"
                              "a b c d e f\n")
dummy_target_text_factor_2 = ("1 2 3 4\n"
                              "1 2 3 4 5 6\n")
dummy_target_text_factored_format = ("Das|a|1 ist|b|2 ein|c|3 Beispieltext.|d|4\n"
                                     "Die|a|1 Factors|b|2 hier|c|3 haben|d|4 keinen|e|5 Sinn.|f|6\n")


def test_translation_factors_dataset():
  """
  Similar to test_translation_dataset(), but using translation factors.
  """
  source_text_per_factor = [dummy_source_text_factor_0, dummy_source_text_factor_1]
  target_text_per_factor = [dummy_target_text_factor_0, dummy_target_text_factor_1, dummy_target_text_factor_2]

  source_vocabulary_names = ["source.vocab.pkl", "source_factor1.vocab.pkl"]
  target_vocabulary_names = ["target.vocab.pkl", "target_factor1.vocab.pkl", "target_factor2.vocab.pkl"]

  source_data_keys = ["data", "source_factor1"]
  target_data_keys = ["classes", "target_factor1", "target_factor2"]

  dummy_dataset = tempfile.mkdtemp()
  source_file_name = os.path.join(dummy_dataset, "source.test.gz")  # testing both zipped and unzipped
  target_file_name = os.path.join(dummy_dataset, "target.test")

  with gzip.open(source_file_name, "wb") as source_file:  # writing in binary format works both in Python 2 and 3
    source_file.write(dummy_source_text_factored_format.encode("utf8"))

  with open(target_file_name, "wb") as target_file:
    target_file.write(dummy_target_text_factored_format.encode("utf8"))

  for postfix in ["", " </S>"]:  # test with and without postfix
    vocabularies, inverse_vocabularies = [], []
    for dummy_text in source_text_per_factor + target_text_per_factor:
      # Append postfix just to have it in the vocabulary
      vocabulary, inverse_vocabulary = create_vocabulary(dummy_text + postfix)
      vocabularies.append(vocabulary)
      inverse_vocabularies.append(inverse_vocabulary)

    vocabulary_names = source_vocabulary_names + target_vocabulary_names
    for index, vocabulary in enumerate(vocabularies):
      with open(os.path.join(dummy_dataset, vocabulary_names[index]), "wb") as vocabulary_file:
        pickle.dump(vocabulary, vocabulary_file)

    translation_dataset = TranslationFactorsDataset(
      path=dummy_dataset, file_postfix="test", factor_separator="|",
      source_factors=source_data_keys[1:], target_factors=target_data_keys[1:],
      source_postfix=postfix, target_postfix=postfix)

    translation_dataset.init_seq_order()
    translation_dataset.load_seqs(0, 10)

    num_seqs = len(dummy_target_text_factored_format.splitlines())
    assert_equal(translation_dataset.num_seqs, num_seqs)

    # Reconstruct the sentences from the word ids for all factors and compare with input.
    data_keys = source_data_keys + target_data_keys
    texts_per_factor = source_text_per_factor + target_text_per_factor
    for index, text in enumerate(texts_per_factor):
      for sequence_index in range(num_seqs):
        word_ids = translation_dataset.get_data(sequence_index, data_keys[index])
        sentence = word_ids_to_sentence(word_ids, inverse_vocabularies[index])
        assert_equal(sentence, text.splitlines()[sequence_index] + postfix)

  shutil.rmtree(dummy_dataset)


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute
